from abc import ABC
from dataclasses import asdict
from smtplib import SMTPException
from typing import List
from uuid import UUID

from asgiref.sync import sync_to_async
from django.contrib.auth import get_user_model
from django.contrib.auth.forms import PasswordChangeForm, SetPasswordForm
from django.core.exceptions import ObjectDoesNotExist
from django.core.signing import BadSignature, SignatureExpired
from django.db import transaction
import strawberry
from strawberry.types import Info

from gqlauth.captcha.models import Captcha as CaptchaModel
from gqlauth.captcha.types_ import CaptchaType
from gqlauth.core.constants import Messages, TokenAction
from gqlauth.core.directives import BaseAuthDirective, IsVerified
from gqlauth.core.exceptions import (
    EmailAlreadyInUse,
    PasswordAlreadySetError,
    TokenScopeError,
    UserAlreadyVerified,
    UserNotVerified,
)
from gqlauth.core.shortcuts import get_user_by_email
from gqlauth.core.types_ import AuthOutput, MutationNormalOutput
from gqlauth.core.utils import (
    get_payload_from_token,
    get_user,
    inject_fields,
    revoke_user_refresh_token,
)
from gqlauth.jwt.types_ import (
    ObtainJSONWebTokenInput,
    ObtainJSONWebTokenType,
    RevokeRefreshTokenType,
    TokenType,
    VerifyTokenInput,
    VerifyTokenType,
)
from gqlauth.models import RefreshToken, UserStatus
from gqlauth.settings import gqlauth_settings as app_settings
from gqlauth.user.forms import (
    EmailForm,
    PasswordLessRegisterForm,
    RegisterForm,
    UpdateAccountForm,
)
from gqlauth.user.helpers import check_captcha, check_secondary_email, confirm_password
from gqlauth.user.signals import user_registered, user_verified

UserModel = get_user_model()


class BaseMixin(ABC):
    directives: List[BaseAuthDirective] = []

    def resolve_mutation(self, *args, **kwargs):
        raise NotImplementedError


class Captcha:
    """
    Creates a brand-new captcha.
    Returns a base64 encoded string of the captcha.
    And uuid representing the captcha id in the database.
    When you will try to log in or register You will
    need submit that uuid With the user input.
    **The captcha will be invoked when the timeout expires**.
    """

    @strawberry.mutation(description=__doc__)
    def field(self) -> CaptchaType:
        return CaptchaModel.create_captcha()

    @strawberry.mutation(description=__doc__)
    @sync_to_async
    def afield(self) -> CaptchaType:
        return CaptchaModel.create_captcha()


class RegisterMixin(BaseMixin):
    """
    Register user with fields defined in the settings.
    If the email field of the user model is part of the
    registration fields (default), check if there is
    no user with that email or as a secondary email.

    If it exists, it does not register the user,
    even if the email field is not defined as unique
    (default of the default django user model).

    When creating the user, it also creates a `UserStatus`
    related to that user, making it possible to track
    if the user is archived, verified and has a secondary
    email.

    Send account verification email.

    If allowed to not verified users login, return token.
    """

    @strawberry.input
    @inject_fields(app_settings.REGISTER_MUTATION_FIELDS)
    class RegisterInput:
        if not app_settings.ALLOW_PASSWORDLESS_REGISTRATION:
            password1: str
            password2: str

        if app_settings.REGISTER_REQUIRE_CAPTCHA:
            identifier: UUID
            userEntry: str

    form = (
        PasswordLessRegisterForm if app_settings.ALLOW_PASSWORDLESS_REGISTRATION else RegisterForm
    )

    @classmethod
    def resolve_mutation(cls, info, input_: RegisterInput) -> MutationNormalOutput:
        if app_settings.LOGIN_REQUIRE_CAPTCHA:
            check_res = check_captcha(input_)
            if check_res != Messages.CAPTCHA_VALID:
                return MutationNormalOutput(success=False, errors={"captcha": check_res})

        try:
            with transaction.atomic():
                f = cls.form(asdict(input_))
                if f.is_valid():
                    email = getattr(input_, "email", False)
                    UserStatus.clean_email(email)
                    user = f.save()
                    if email:
                        send_activation = app_settings.SEND_ACTIVATION_EMAIL is True
                        send_password_set = (
                            app_settings.ALLOW_PASSWORDLESS_REGISTRATION is True
                            and app_settings.SEND_PASSWORD_SET_EMAIL is True
                        )
                        if send_activation:
                            user.status.send_activation_email(info)

                        if send_password_set:
                            user.status.send_password_set_email(info)

                    user_registered.send(sender=cls, user=user)
                    return MutationNormalOutput(success=True)
                else:
                    return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except EmailAlreadyInUse:
            return MutationNormalOutput(
                success=False,
                # if the email was set as a secondary email,
                # the RegisterForm will not catch it,
                # so we need to run UserStatus.clean_email(email)
                errors={UserModel.EMAIL_FIELD: Messages.EMAIL_IN_USE},
            )
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)


class VerifyAccountMixin(BaseMixin):
    """
    Verify user account.

    Receive the token that was sent by email.
    If the token is valid, make the user verified
    by making the `user.status.verified` field true.
    """

    @strawberry.input
    class VerifyAccountInput:
        token: str

    @classmethod
    def resolve_mutation(cls, info: Info, input_: VerifyAccountInput) -> MutationNormalOutput:
        try:
            UserStatus.verify(input_.token)
            return MutationNormalOutput(success=True)
        except UserAlreadyVerified:
            return MutationNormalOutput(success=False, errors=Messages.ALREADY_VERIFIED)
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)


class VerifySecondaryEmailMixin(BaseMixin):
    """
    Verify user secondary email.

    Receive the token that was sent by email.
    User is already verified when using this mutation.

    If the token is valid, add the secondary email
    to `user.status.secondary_email` field.

    Note that until the secondary email is verified,
    it has not been saved anywhere beyond the token,
    so it can still be used to create a new account.
    After being verified, it will no longer be available.
    """

    @strawberry.input
    class VerifySecondaryEmailInput:
        token: str

    @classmethod
    def resolve_mutation(cls, _, input_: VerifySecondaryEmailInput) -> MutationNormalOutput:
        try:
            token = input_.token
            UserStatus.verify_secondary_email(token)
            return MutationNormalOutput(success=True)
        except EmailAlreadyInUse:
            # while the token was sent and the user haven't
            # verified, the email was free. If other account
            # was created with it, it is already in use.
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_IN_USE)
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)


class ResendActivationEmailMixin(BaseMixin):
    """
    Sends activation email.

    It is called resend because theoretically
    the first activation email was sent when
    the user registered.

    If there is no user with the requested email,
    a successful response is returned.
    """

    @strawberry.input
    class ResendActivationEmailInput:
        email: str

    @classmethod
    def resolve_mutation(cls, info, input_: ResendActivationEmailInput) -> MutationNormalOutput:
        try:
            email = input_.email
            f = EmailForm({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                user.status.resend_activation_email(info)
                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except ObjectDoesNotExist:
            return MutationNormalOutput(success=True)  # even if user is not registered
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)
        except UserAlreadyVerified:
            return MutationNormalOutput(success=False, errors={"email": Messages.ALREADY_VERIFIED})


class SendPasswordResetEmailMixin(BaseMixin):
    """
    Send password reset email.

    For non verified users, send an activation
    email instead.

    Accepts both primary and secondary email.

    If there is no user with the requested email,
    a successful response is returned.
    """

    @strawberry.input
    class SendPasswordResetEmailInput:
        email: str

    @classmethod
    def resolve_mutation(cls, info, input_: SendPasswordResetEmailInput) -> MutationNormalOutput:
        try:
            email = input_.email
            f = EmailForm({"email": email})
            if f.is_valid():
                user = get_user_by_email(email)
                user.status.send_password_reset_email(info, [email])
                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except ObjectDoesNotExist:
            return MutationNormalOutput(success=True)  # even if user is not registered
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)
        except UserNotVerified:
            user = get_user_by_email(input_.email)
            try:
                user.status.resend_activation_email(info)
                return MutationNormalOutput(
                    success=False,
                    errors={"email": Messages.NOT_VERIFIED_PASSWORD_RESET},
                )
            except SMTPException:
                return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)


class PasswordResetMixin(BaseMixin):
    """
    Change user password without old password.

    Receive the token that was sent by email.

    If token and new passwords are valid, update
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    @strawberry.input
    class PasswordResetInput:
        token: str
        new_password1: str
        new_password2: str

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, _, input_: PasswordResetInput) -> MutationNormalOutput:
        try:
            payload = get_payload_from_token(
                input_.token,
                TokenAction.PASSWORD_RESET,
                app_settings.EXPIRATION_PASSWORD_RESET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)

            f = cls.form(user, asdict(input_))
            if f.is_valid():
                revoke_user_refresh_token(user)
                user = f.save()
                if user.status.verified is False:
                    user.status.verified = True
                    user.status.save(update_fields=["verified"])
                    user_verified.send(sender=cls, user=user)

                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)


class PasswordSetMixin(BaseMixin):
    """
    Set user password - for password-less registration

    Receive the token that was sent by email.

    If token and new passwords are valid, set
    user password and in case of using refresh
    tokens, revoke all of them.

    Also, if user has not been verified yet, verify it.
    """

    @strawberry.input
    class PasswordSetInput:
        token: str
        new_password1: str
        new_password2: str

    form = SetPasswordForm

    @classmethod
    def resolve_mutation(cls, _, input_: PasswordSetInput) -> MutationNormalOutput:
        try:
            token = input_.token
            payload = get_payload_from_token(
                token,
                TokenAction.PASSWORD_SET,
                app_settings.EXPIRATION_PASSWORD_SET_TOKEN,
            )
            user = UserModel._default_manager.get(**payload)
            f = cls.form(user, asdict(input_))
            if f.is_valid():
                # Check if user has already set a password
                if user.has_usable_password():
                    raise PasswordAlreadySetError
                revoke_user_refresh_token(user)
                user = f.save()

                if user.status.verified is False:
                    user.status.verified = True
                    user.status.save(update_fields=["verified"])

                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except SignatureExpired:
            return MutationNormalOutput(success=False, errors=Messages.EXPIRED_TOKEN)
        except (BadSignature, TokenScopeError):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_TOKEN)
        except PasswordAlreadySetError:
            return MutationNormalOutput(success=False, errors=Messages.PASSWORD_ALREADY_SET)


class ObtainJSONWebTokenMixin(BaseMixin):
    """
    Obtain JSON web token for given user.

    Allow to perform login with different fields,
    and secondary email if set. The fields are
    defined on settings.

    Not verified users can log in by default. This
    can be changes on settings.

    If user is archived, make it unarchived and
    return `unarchiving=True` on OutputBase.
    """

    @classmethod
    def resolve_mutation(cls, info, input_: ObtainJSONWebTokenInput) -> ObtainJSONWebTokenType:
        if app_settings.LOGIN_REQUIRE_CAPTCHA:
            check_res = check_captcha(input_)
            if check_res != Messages.CAPTCHA_VALID:
                return ObtainJSONWebTokenType(success=False, errors={"captcha": check_res})

        return ObtainJSONWebTokenType.authenticate(info, input_)


class ArchiveOrDeleteMixin(BaseMixin):
    @strawberry.input
    class ArchiveOrDeleteMixinInput:
        password: str

    directives = [
        IsVerified(),
    ]

    @classmethod
    def resolve_mutation(
        cls, info, input_: ArchiveOrDeleteMixinInput
    ) -> AuthOutput[MutationNormalOutput]:
        user = get_user(info)
        if error := confirm_password(user, input_):
            return AuthOutput(node=error)
        cls.resolve_action(user)
        return AuthOutput(node=MutationNormalOutput(success=True))


class ArchiveAccountMixin(ArchiveOrDeleteMixin):
    """
    Archive account and revoke refresh tokens.
    User must be verified and confirm password.
    """

    @classmethod
    def resolve_action(cls, user):
        UserStatus.archive(user)
        revoke_user_refresh_token(user=user)


class DeleteAccountMixin(ArchiveOrDeleteMixin):
    """
    Delete account permanently or make `user.is_active=False`.

    The behavior is defined on settings.
    Anyway user refresh tokens are revoked.

    User must be verified and confirm password.
    """

    @classmethod
    def resolve_action(cls, user):
        if app_settings.ALLOW_DELETE_ACCOUNT:
            revoke_user_refresh_token(user=user)
            user.delete()


class PasswordChangeMixin(BaseMixin):
    """
    Change account password when user knows the old password.

    A new token and refresh token are sent. User must be verified.
    """

    @strawberry.input
    class PasswordChangeInput:
        old_password: str
        new_password1: str
        new_password2: str

    form = PasswordChangeForm
    directives = [
        IsVerified(),
    ]

    @classmethod
    def resolve_mutation(cls, info: Info, input_: PasswordChangeInput) -> ObtainJSONWebTokenType:
        user = get_user(info)
        if error := confirm_password(user, input_):
            return ObtainJSONWebTokenType(**asdict(error))

        args = asdict(input_)
        f = cls.form(user, args)
        if f.is_valid():
            revoke_user_refresh_token(user)
            user = f.save()
            return ObtainJSONWebTokenType.from_user(info, user)
        else:
            return ObtainJSONWebTokenType(success=False, errors=f.errors.get_json_data())


class UpdateAccountMixin(BaseMixin):
    """
    Update user model fields, defined on settings.

    User must be verified.
    """

    @strawberry.input
    @inject_fields(app_settings.UPDATE_MUTATION_FIELDS)
    class UpdateAccountInput:
        ...

    form = UpdateAccountForm
    directives = [
        IsVerified(),
    ]

    @classmethod
    def resolve_mutation(cls, info, input_: UpdateAccountInput) -> AuthOutput[MutationNormalOutput]:
        user = get_user(info)
        f = cls.form(
            asdict(input_),
            instance=user,
        )
        if f.is_valid():
            f.save()
            return AuthOutput(node=MutationNormalOutput(success=True))
        else:
            return AuthOutput(
                node=MutationNormalOutput(success=False, errors=f.errors.get_json_data())
            )


class VerifyTokenMixin(BaseMixin):
    """
    ### Checks if a token is not expired and correct.
    *Note that this is not for refresh tokens.*
    """

    @classmethod
    def resolve_mutation(cls, _: Info, input_: VerifyTokenInput) -> VerifyTokenType:
        return VerifyTokenType.from_token(input_)


class RefreshTokenMixin(BaseMixin):
    """
    ### refreshToken to generate a new login token:
    *Use this only if `JWT_LONG_RUNNING_REFRESH_TOKEN` is True*

    using the refresh-token you already got during authorization, and
    obtain a brand-new token (and possibly a new refresh token if you revoked the previous).
    This is an alternative to log in when your token expired.
    """

    @strawberry.input(
        description="If `revoke_refresh_token` is true,"
        " revokes to refresh token an creates a new one."
    )
    class RefreshTokenInput:
        refresh_token: str
        revoke_refresh_token: bool = strawberry.field(
            description="revokes the previous refresh token, and will generate new one.",
            default=False,
        )

    @classmethod
    def resolve_mutation(cls, info, input_: RefreshTokenInput) -> ObtainJSONWebTokenType:
        try:
            res = RefreshToken.objects.get(token=input_.refresh_token)
        except RefreshToken.DoesNotExist:
            return ObtainJSONWebTokenType(success=False, errors=Messages.INVALID_TOKEN)
        user = res.user
        if res.is_expired_():
            return ObtainJSONWebTokenType(success=False, errors=Messages.EXPIRED_TOKEN)
        ret = ObtainJSONWebTokenType(
            success=True, token=TokenType.from_user(info, user), refresh_token=res
        )
        if input_.revoke_refresh_token:
            res.revoke()
            ret.refresh_token = RefreshToken.from_user(user)
        return ret


class RevokeTokenMixin(BaseMixin):
    """
    ### Suspends a refresh token.
    *token must exist to be revoked.*
    """

    @strawberry.input
    class RevokeTokenInput:
        refresh_token: str

    @classmethod
    def resolve_mutation(cls, _: Info, input_: RevokeTokenInput) -> RevokeRefreshTokenType:
        try:
            refresh_token = RefreshToken.objects.get(
                token=input_.refresh_token, revoked__isnull=True
            )
            refresh_token.revoke()
            return RevokeRefreshTokenType(success=True, refresh_token=refresh_token)
        except RefreshToken.DoesNotExist:
            return RevokeRefreshTokenType(success=False, errors=Messages.INVALID_TOKEN)


class SendSecondaryEmailActivationMixin(BaseMixin):
    """
    Send activation to secondary email.

    User must be verified and confirm password.
    """

    directives = [
        IsVerified(),
    ]

    @strawberry.input
    class SendSecondaryEmailActivationInput:
        password: str

    @classmethod
    def resolve_mutation(
        cls, info, input_: SendSecondaryEmailActivationInput
    ) -> MutationNormalOutput:
        user = get_user(info)
        if error := confirm_password(user, input_):
            return error
        try:
            email = input_.get("email")
            f = EmailForm({"email": email})
            if f.is_valid():
                user.status.send_secondary_email_activation(info, email)
                return MutationNormalOutput(success=True)
            return MutationNormalOutput(success=False, errors=f.errors.get_json_data())
        except EmailAlreadyInUse:
            # while the token was sent and the user haven't verified,
            # the email was free. If other account was created with if
            # it is already in use
            return MutationNormalOutput(success=False, errors={"email": Messages.EMAIL_IN_USE})
        except SMTPException:
            return MutationNormalOutput(success=False, errors=Messages.EMAIL_FAIL)


class SwapEmailsMixin(BaseMixin):
    """
    Swap between primary and secondary emails.

    Require password confirmation.
    """

    @strawberry.input
    class SwapEmailsInput:
        password: str

    @classmethod
    def resolve_mutation(cls, info: Info, input_: SwapEmailsInput) -> MutationNormalOutput:
        user = get_user(info)
        if not user.check_password(input_.password):
            return MutationNormalOutput(success=False, errors=Messages.INVALID_PASSWORD)
        if error := check_secondary_email(user):
            return error
        user.status.swap_emails()
        return MutationNormalOutput(success=True)


class RemoveSecondaryEmailMixin(BaseMixin):
    """
    Remove user secondary email.

    Require password confirmation.
    """

    @strawberry.input
    class RemoveSecondaryEmailInput:
        password: str

    @classmethod
    def resolve_mutation(
        cls, info: Info, input_: RemoveSecondaryEmailInput
    ) -> MutationNormalOutput:
        user = get_user(info)

        if error := confirm_password(user, input_):
            return error

        if error := check_secondary_email(user):
            return error

        user.status.remove_secondary_email()
        return MutationNormalOutput(success=True)



from abc import ABC, abstractmethod
from typing import Any, List, Optional

from django.contrib.auth import get_user_model
from django.utils.translation import gettext as _
import strawberry
from strawberry.schema_directive import Location
from strawberry.types import Info

from gqlauth.core.types_ import ErrorMessage, GqlAuthError
from gqlauth.core.utils import get_status

USER_MODEL = get_user_model()

__all__ = ["BaseAuthDirective", "IsAuthenticated", "HasPermission", "IsVerified"]


class BaseAuthDirective(ABC):
    @abstractmethod
    def resolve_permission(
        self, user: USER_MODEL, source: Any, info: Info, args: list, kwargs: dict
    ) -> Optional[ErrorMessage]:
        ...


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires authentication",
)
class IsAuthenticated(BaseAuthDirective):
    def resolve_permission(self, user: USER_MODEL, source: Any, info: Info, *args, **kwargs):
        if not user.is_authenticated:
            return ErrorMessage(code=GqlAuthError.UNAUTHENTICATED)
        return None


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires the user to be verified",
)
class IsVerified(BaseAuthDirective):
    def resolve_permission(self, user: USER_MODEL, source: Any, info: Info, *args, **kwargs):
        if (status := get_status(user)) and status.verified:
            return None
        return ErrorMessage(code=GqlAuthError.NOT_VERIFIED)


@strawberry.schema_directive(
    locations=[
        Location.FIELD_DEFINITION,
    ],
    description="This field requires a certain permissions to be resolved.",
)
class HasPermission(BaseAuthDirective):
    permissions: strawberry.Private[List[str]]

    def resolve_permission(self, user: USER_MODEL, source: Any, info: Info, *args, **kwargs):
        for permission in self.permissions:
            if not user.has_perm(permission):
                return ErrorMessage(
                    code=GqlAuthError.NO_SUFFICIENT_PERMISSIONS,
                    message=_(
                        f"User {user.first_name or getattr(user, user.USERNAME_FIELD, None)}, has not sufficient permissions for {info.path.key}"
                    ),
                )
        return None



from typing import Any, Awaitable, Callable, Dict, List, Union

from asgiref.sync import sync_to_async
from django.contrib.auth import get_user_model
from jwt import PyJWTError
from strawberry import UNSET
from strawberry.field import StrawberryField
from strawberry.types import Info
from strawberry_django import django_resolver
from strawberry_django.fields.field import StrawberryDjangoField
from strawberry_django.utils import is_async

from gqlauth.core.directives import BaseAuthDirective
from gqlauth.core.exceptions import TokenExpired
from gqlauth.core.types_ import AuthOutput, ErrorMessage, GqlAuthError
from gqlauth.core.utils import get_user
from gqlauth.jwt.types_ import TokenType
from gqlauth.settings import gqlauth_settings as app_settings

__all__ = ["GqlAuthRootField", "field"]

USER_MODEL = get_user_model()


class GqlAuthRootField(StrawberryDjangoField):
    def _resolve(
        self, token: TokenType, source: Any, info: Info, args: List[Any], kwargs: Dict[str, Any]
    ):
        user = token.get_user_instance()
        info.context.user = user
        return super().get_result(source, info, args, kwargs)

    def get_result(
        self, source: Any, info: Info, args: List[Any], kwargs: Dict[str, Any]
    ) -> Union[Awaitable[Any], Any]:
        token = app_settings.JWT_TOKEN_FINDER(info)
        try:
            token_type = TokenType.from_token(token)
        except PyJWTError:  # raised by python-jwt
            return AuthOutput(error=ErrorMessage(code=GqlAuthError.INVALID_TOKEN))

        except TokenExpired:
            return AuthOutput(error=ErrorMessage(code=GqlAuthError.EXPIRED_TOKEN))

        if is_async():
            return sync_to_async(self._resolve)(token_type, source, info, args, kwargs)
        else:
            return self._resolve(token_type, source, info, args, kwargs)


class GqlAuthField(StrawberryDjangoField):
    def _resolve(self, source, info, args, kwargs):
        user = get_user(info)
        for directive in self.directives:

            if isinstance(directive, BaseAuthDirective) and (
                error := directive.resolve_permission(user, source, info, args, kwargs)
            ):
                return AuthOutput(error=error)
        return super().get_result(source, info, args, kwargs)

    def get_result(self, source, info, args, kwargs):
        if is_async():
            return sync_to_async(self._resolve)(source, info, args, kwargs)
        return self._resolve(source, info, args, kwargs)


def field(
    resolver=None,
    *,
    name=None,
    field_name=None,
    filters=UNSET,
    default=UNSET,
    directives: List[BaseAuthDirective] = None,
    **kwargs,
) -> Union[StrawberryField, Callable[..., StrawberryField]]:
    field_ = GqlAuthField(
        python_name=None,
        graphql_name=name,
        filters=filters,
        django_name=field_name,
        default=default,
        directives=directives,
        **kwargs,
    )
    if resolver:
        resolver = django_resolver(resolver)
        field_ = field_(resolver)
    return field_



import dataclasses
from datetime import datetime
from typing import Optional
from uuid import UUID

from django.conf import settings
from django.contrib.auth import authenticate, get_user_model
from django.core.exceptions import PermissionDenied
import strawberry
from strawberry import auto
from strawberry.types import Info
import strawberry_django

from gqlauth.core.constants import Messages
from gqlauth.core.exceptions import TokenExpired
from gqlauth.core.interfaces import OutputInterface
from gqlauth.core.scalars import ExpectedErrorType
from gqlauth.core.utils import inject_fields
from gqlauth.models import RefreshToken
from gqlauth.user.types_ import UserType

app_settings = settings.GQL_AUTH

USER_MODEL = get_user_model()


@strawberry_django.type(
    model=RefreshToken,
    description="""
Refresh token can be used to obtain a new token instead of log in again
when the token expires.

*This is only used if `JWT_LONG_RUNNING_REFRESH_TOKEN` is set to True.*
""",
)
class RefreshTokenType:
    token: auto = strawberry_django.field(
        description="randomly generated token that is attached to a FK user."
    )
    created: auto
    revoked: auto

    @strawberry_django.field
    def expires_at(self: RefreshToken) -> datetime:
        return self.expires_at_()

    @strawberry_django.field
    def is_expired(self: RefreshToken) -> bool:
        return self.is_expired_()


@strawberry.type(
    description="""
the data that was used to create the token.
"""
)
@inject_fields(
    {
        app_settings.JWT_PAYLOAD_PK,
    }
)
class TokenPayloadType:
    origIat: datetime = strawberry.field(
        description="when the token was created", default_factory=datetime.utcnow
    )
    exp: datetime = strawberry.field(description="when the token will be expired", default=None)

    def __post_init__(self):
        if not self.exp:
            self.exp = self.origIat + app_settings.JWT_EXPIRATION_DELTA

    def as_dict(self):
        ret = dataclasses.asdict(self)
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if isinstance(value, datetime):
                ret[field.name] = value.strftime(app_settings.JWT_TIME_FORMAT)
        return ret

    @classmethod
    def from_dict(cls, data: dict) -> "TokenPayloadType":
        for field in dataclasses.fields(cls):
            value = data[field.name]
            if isinstance(value, str) and field.type is datetime:
                data[field.name] = datetime.strptime(value, app_settings.JWT_TIME_FORMAT)
        return cls(**data)


@strawberry.type(
    description="""
encapsulates the token with the payload that was used to create the token.
"""
)
class TokenType:
    payload: TokenPayloadType
    token: str = strawberry.field(description="The encoded payload, namely a token.")

    def is_expired(self):
        return self.payload.exp < (datetime.utcnow())

    @classmethod
    def from_user(cls, info: Info, user: USER_MODEL) -> "TokenType":
        return app_settings.JWT_PAYLOAD_HANDLER(info, user)

    @classmethod
    def from_token(cls, token: str) -> "TokenType":
        """
        might raise TokenExpired
        """
        token_type: TokenType = app_settings.JWT_DECODE_HANDLER(token)
        if token_type.is_expired():
            raise TokenExpired
        return token_type

    def get_user_instance(self) -> USER_MODEL:
        """
        might raise not existed exception.
        """
        pk_name = app_settings.JWT_PAYLOAD_PK.python_name
        query = {pk_name: getattr(self.payload, pk_name)}
        return USER_MODEL.objects.get(**query)


@strawberry.input
@inject_fields(app_settings.LOGIN_FIELDS)
class ObtainJSONWebTokenInput:
    password: str
    if app_settings.LOGIN_REQUIRE_CAPTCHA:
        identifier: UUID
        userEntry: str


@strawberry.type(
    description="""
    encapsulates token data, and refresh token data if `JWT_LONG_RUNNING_REFRESH_TOKEN` is on.
    with an output interface.
    """
)
class ObtainJSONWebTokenType(OutputInterface):
    success: bool
    user: Optional[UserType] = None
    token: Optional[TokenType] = None
    if app_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
        refresh_token: Optional[RefreshTokenType] = None
    errors: Optional[ExpectedErrorType] = None

    @classmethod
    def from_user(cls, info: Info, user: USER_MODEL) -> "ObtainJSONWebTokenType":
        """
        creates a new token and possibly a new refresh token based on the user.
        *call this method only for trusted users.*
        """
        ret = ObtainJSONWebTokenType(success=True, user=user, token=TokenType.from_user(info, user))
        if app_settings.JWT_LONG_RUNNING_REFRESH_TOKEN:
            ret.refresh_token = RefreshToken.from_user(user)
        return ret

    @classmethod
    def authenticate(cls, info: Info, input_: ObtainJSONWebTokenInput) -> "ObtainJSONWebTokenType":
        """
        return `ObtainJSONWebTokenType`.
        authenticates against django authentication backends.

        *creates a new token and possibly a refresh token.*
        """
        args = {
            USER_MODEL.USERNAME_FIELD: getattr(input_, USER_MODEL.USERNAME_FIELD),
            "password": input_.password,
        }
        try:
            # authenticate against django authentication backends.
            if not (user := authenticate(info.context.request, **args)):
                return ObtainJSONWebTokenType(success=False, errors=Messages.INVALID_CREDENTIALS)

            # gqlauth logic
            if user.status.archived is True:  # un-archive on login
                from gqlauth.models import UserStatus

                UserStatus.unarchive(user)
            if user.status.verified or app_settings.ALLOW_LOGIN_NOT_VERIFIED:
                # successful login.
                return ObtainJSONWebTokenType.from_user(info, user)
            else:
                return ObtainJSONWebTokenType(success=False, errors=Messages.NOT_VERIFIED)

        except PermissionDenied:
            # one of the authentication backends rejected the user.
            return ObtainJSONWebTokenType(success=False, errors=Messages.UNAUTHENTICATED)

        except TokenExpired:
            return ObtainJSONWebTokenType(success=False, errors=Messages.EXPIRED_TOKEN)


@strawberry.input
class VerifyTokenInput:
    token: str


@strawberry.type
class VerifyTokenType(OutputInterface):
    success: bool
    token: Optional[TokenType] = None
    user: Optional[UserType] = None
    errors: Optional[ExpectedErrorType] = None

    @classmethod
    def from_token(cls, token_input: VerifyTokenInput) -> "VerifyTokenType":
        try:
            token_type = TokenType.from_token(token_input.token)
            user = token_type.get_user_instance()
        except USER_MODEL.DoesNotExist:
            return VerifyTokenType(success=False, errors=Messages.INVALID_CREDENTIALS)
        except TokenExpired:
            return VerifyTokenType(success=False, errors=Messages.EXPIRED_TOKEN)

        else:
            return VerifyTokenType(token=token_type, user=user, success=True)


@strawberry.type
class RevokeRefreshTokenType:
    success: bool
    refresh_token: Optional[RefreshTokenType] = None
    errors: Optional[ExpectedErrorType] = None



import contextlib
import inspect
import typing
from typing import Dict, Iterable, Union

from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core import signing
from strawberry.field import StrawberryField
from strawberry.types import Info
from strawberry.utils.str_converters import to_camel_case

from gqlauth.core.exceptions import TokenScopeError

if typing.TYPE_CHECKING:
    from gqlauth.models import UserStatus

USER_MODEL = get_user_model()


def hide_args_kwargs(field):
    sig = inspect.signature(field)
    cleared = tuple(p for p in sig.parameters.values() if p.name not in ("kwargs", "args"))
    field.__signature__ = inspect.signature(field).replace(parameters=(cleared))
    return field


def isiterable(value):
    try:
        iter(value)
    except TypeError:
        return False
    return True


def camelize(data):
    if isinstance(data, dict):
        return {to_camel_case(k): camelize(v) for k, v in data.items()}
    if isiterable(data) and not isinstance(data, str):
        return [camelize(d) for d in data]
    return data


def list_to_dict(lst: [str]):
    """takes list of string and creates a dict with str as their values"""
    new_dict = {}
    for item in lst:
        new_dict[item] = str
    return new_dict


def get_request(info: Info):
    if hasattr(info.context, "user"):
        return info.context
    return info.context.request


def get_info(args: tuple) -> typing.Optional[Info]:
    for arg in args:
        if isinstance(arg, Info):
            return arg
    return None


def get_status(user: Union[USER_MODEL, AnonymousUser]) -> typing.Optional["UserStatus"]:
    from gqlauth.models import UserStatus

    if status := getattr(user, "status", False):
        assert isinstance(status, UserStatus)
        return status
    return None


def get_user(info: Info) -> Union[USER_MODEL, AnonymousUser]:
    if user := getattr(info.context, "user", None):  # noqa: B009
        assert isinstance(user, USER_MODEL)
        return user
    return AnonymousUser()


def get_token(user, action, **kwargs):
    username = user.get_username()
    if hasattr(username, "pk"):
        username = username.pk
    payload = {user.USERNAME_FIELD: username, "action": action}
    if kwargs:
        payload.update(**kwargs)
    token = signing.dumps(payload)
    return token


def get_payload_from_token(token, action, exp=None):
    payload = signing.loads(token, max_age=exp)
    _action = payload.pop("action")
    if _action != action:
        raise TokenScopeError
    return payload


def revoke_user_refresh_token(user):
    refresh_tokens = user.refresh_tokens.all()
    for refresh_token in refresh_tokens:
        with contextlib.suppress(Exception):
            refresh_token.revoke()


def fields_names(strawberry_fields: Iterable[StrawberryField]):
    return [field.python_name for field in strawberry_fields]


def normalize_fields(dict_or_list, extra_list_or_dict):
    """
    helper merge settings defined filed
    with default str type
    """
    if not isinstance(extra_list_or_dict, dict):
        extra_list_or_dict = list_to_dict(extra_list_or_dict)
    if not isinstance(dict_or_list, dict):
        dict_or_list = list_to_dict(dict_or_list)
    dict_or_list.update(extra_list_or_dict)

    return dict_or_list


def inject_fields(fields: typing.Iterable[StrawberryField], annotations_only=False):
    def wrapped(cls: type):
        # python 3.8 compat:
        if not hasattr(cls, "__annotations__"):
            cls.__annotations__ = {}

        for field in fields:
            if not field.name:
                continue
            if not annotations_only:
                setattr(cls, field.name, field)
            cls.__annotations__[field.name] = field.type_annotation.annotation
        return cls

    return wrapped


def inject_arguments(args: Dict[str, type]):
    """
    injects arguments to the decorated resolver.
    :param args: `dict[name, type]` of arguments to be injected.,
    """

    def wrapped(fn):
        sig = inspect.signature(fn)
        params = {
            inspect.Parameter(name, inspect.Parameter.POSITIONAL_OR_KEYWORD, annotation=type_)
            for name, type_ in args.items()
        }
        params.update(sig.parameters.values())
        fn.__signature__ = inspect.signature(fn).replace(parameters=params)
        return fn

    return wrapped


def is_optional(field):
    """
    whether strawberry field is optional or not
    """
    return typing.get_origin(field) is Union and type(None) in typing.get_args(field)



from typing import get_args
import warnings

from django.conf import settings as django_settings
from django.utils.module_loading import import_string

from gqlauth.settings_type import DjangoSetting, GqlAuthSettings, ImportString

gqlauth_settings: GqlAuthSettings = None

if user_settings := getattr(django_settings, "GQL_AUTH", False):
    if isinstance(user_settings, GqlAuthSettings):
        gqlauth_settings = user_settings

    else:
        raise Exception(
            f"GQL_AUTH settings should be of type "
            f"{GqlAuthSettings}, but you provided {type(user_settings)}"
        )

else:
    warnings.warn("You have not provided any custom gql auth settings falling back to defaults")
    gqlauth_settings = GqlAuthSettings()

defaults = GqlAuthSettings()
# retain django_settings
for name, setting in gqlauth_settings.__dataclass_fields__.items():
    value = getattr(gqlauth_settings, name)
    if setting.type is DjangoSetting and value is getattr(defaults, name):
        setattr(gqlauth_settings, name, value())
    elif ImportString in get_args(setting.type) and isinstance(value, str):
        setattr(gqlauth_settings, name, import_string(value))

del defaults



from typing import TYPE_CHECKING, Optional, Union

from django.contrib.auth import get_user_model

from gqlauth.captcha.models import Captcha as CaptchaModel
from gqlauth.core.constants import Messages
from gqlauth.core.types_ import MutationNormalOutput
from gqlauth.core.utils import get_status

if TYPE_CHECKING:
    from gqlauth.user.resolvers import ObtainJSONWebTokenInput, RegisterMixin

USER_MODEL = get_user_model()


def confirm_password(user: USER_MODEL, input_) -> Optional[MutationNormalOutput]:
    if password := getattr(input_, "password", False):
        password_arg = "password"
    else:
        password = input_.old_password
        password_arg = "oldPassword"

    if user.check_password(password):
        return None
    errors = {password_arg: Messages.INVALID_PASSWORD}
    return MutationNormalOutput(success=False, errors=errors)


def check_secondary_email(user: USER_MODEL) -> Optional[MutationNormalOutput]:
    if (status := get_status(user)) and status.secondary_email:
        return None
    return MutationNormalOutput(success=False, errors=Messages.SECONDARY_EMAIL_REQUIRED)


def check_captcha(input_: Union["RegisterMixin.RegisterInput", "ObtainJSONWebTokenInput"]):
    uuid = input_.identifier
    try:
        obj = CaptchaModel.objects.get(uuid=uuid)
    except CaptchaModel.DoesNotExist:
        return Messages.CAPTCHA_EXPIRED
    return obj.validate(input_.userEntry)



from enum import Enum
from typing import Generic, Optional, TypeVar

from django.utils.translation import gettext as _
import strawberry

from gqlauth.core.scalars import ExpectedErrorType

T = TypeVar("T")


@strawberry.type
class MutationNormalOutput:
    success: bool
    errors: Optional[ExpectedErrorType] = None


@strawberry.type
class ErrorMessage:
    code: "GqlAuthError" = None
    message: str = None

    def __post_init__(self):
        if not self.message:
            assert isinstance(self.code.value, str)
            self.message = _(self.code.value)
        self.message = _(self.message)


@strawberry.type
class AuthOutput(Generic[T]):
    node: Optional[T] = None
    error: Optional[ErrorMessage] = None
    success: bool = False

    def __post_init__(self):
        if self.node:
            assert not self.error
            self.success = True


@strawberry.enum
class GqlAuthError(Enum):
    UNAUTHENTICATED = "Unauthenticated."
    INVALID_TOKEN = "Invalid token."
    EXPIRED_TOKEN = "Expired token."
    NO_SUFFICIENT_PERMISSIONS = "Permissions found could not satisfy the required permissions."
    NOT_VERIFIED = "Please verify your account."



from django.apps import AppConfig


class GqlAuthConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "gqlauth"
    verbose_name = "GraphQL Auth"

    def ready(self):
        pass



from uuid import UUID

import strawberry
import strawberry_django

from gqlauth.captcha import models
from gqlauth.core.scalars import image as Image


@strawberry_django.type(model=models.Captcha)
class CaptchaType:
    uuid: UUID
    image: strawberry.auto

    @strawberry.django.field(description="returns the b64 encoded image.")
    def pil_image(self) -> Image:
        return self.as_bytes()



from typing import Optional, TypeVar

import strawberry

from gqlauth.core.scalars import ExpectedErrorType

T = TypeVar("T")


@strawberry.interface
class OutputInterface:
    success: bool
    errors: Optional[ExpectedErrorType]



from django.contrib.auth import get_user_model
from strawberry.field import StrawberryField
from strawberry.types import Info

import gqlauth
from gqlauth.core.utils import hide_args_kwargs, inject_arguments
from gqlauth.user.resolvers import BaseMixin

UserModel = get_user_model()


class ArgMixin:
    field: StrawberryField

    def __init_subclass__(cls: BaseMixin, **kwargs):
        input_type = cls.resolve_mutation.__annotations__["input_"]
        return_type = cls.resolve_mutation.__annotations__["return"]

        @gqlauth.core.field_.field(description=cls.__doc__, directives=cls.directives)
        @inject_arguments(input_type.__annotations__)
        @hide_args_kwargs
        def field(info: Info, **kwargs) -> return_type:
            return cls.resolve_mutation(info, input_type(**kwargs))

        cls.field = field


class RelayMixin:
    field: StrawberryField
    afield: StrawberryField

    def __init_subclass__(cls: BaseMixin, **kwargs):
        input_type = cls.resolve_mutation.__annotations__["input_"]
        return_type = cls.resolve_mutation.__annotations__["return"]

        @gqlauth.core.field_.field(description=cls.__doc__, directives=cls.directives)
        @hide_args_kwargs
        def field(info: Info, input: input_type) -> return_type:
            return cls.resolve_mutation(info, input)

        cls.field = field
